/**************************************************************************/
/*                                                                        */
/*   Program Name :  pfind                                                */
/*                                                                        */
/*      Version:  1.10  2018-01-08                                        */
/*                                                                        */
/*      General purpose pattern finding                                   */
/*                                                                        */
/*                                                                        */
/*      Copyright (c) 2007-2018 T.Kondo/NICT All Right Reserved           */
/*                                                                        */
/**************************************************************************/
/*

   Ver. 2007-03-04  Start development
        2013-03-06  default pattern changed to VSSP32
        2014-04-16  update to handle very big data
        2018-01-08  update to control output 


      usage
             pfind fname HDpat1 HDpat2
       
              where  fname -- file name
			         HDpat1  ------ corrupted header pattern 1 in 8 digit Hexs or '?'
                                    where '?' is a wild card for one Hex.
                                    (ex. FFFFFF??) (default is 'FFFFFF??')
                     HDpat2  ------ corrupted header pattern 3 in 8 digit Hexs or '?'
                                    where '?' is a wild card for one Hex.
                                    (ex. 8B????FF) (default is '8B????FF')

									as for VSSP set HDpat1 as 'FFFFFF??' and HDpat2 as '8B????FF'
									as for VSSP32 set HDpat1 as 'FFFF????' and HDpat2 as '8C??????'

*/
#define PROGRAM_NAME "pfind"
#define PROGRAM_VERSION "Ver. 2018-01-08"
#define _CRT_SECURE_NO_WARNINGS     // to supress warning messages

#define VSSP32_MODE         // set VSSP32 mode

/* set serach default pattern */
#ifdef VSSP32_MODE 
   #define DFLT_HDPAT1 "FFFF????"    // for VSSP32
   #define DFLT_HDPAT2 "8C??????"    // for VSSP32
#else
   #define DFLT_HDPAT1 "FFFFFF??"    // for VSSP typical destroyed pattern
   #define DFLT_HDPAT2 "8B????FF"    // for VSSP typical destroyed pattern
#endif

#include <stdio.h>
#include <signal.h> 
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef __CYGWIN__
#ifndef linux
#define linux
#endif
#endif

#ifdef linux
#include <sys/time.h>
#endif

#ifndef WIN32
#include <unistd.h>
#endif


#ifdef WIN32
#ifndef _FORWIN32SET_
#define _FORWIN32SET_
typedef unsigned int    u_int32_t;
typedef unsigned int    uint32_t;
typedef unsigned short int    u_int16_t;
typedef unsigned short int    uint16_t;
typedef unsigned char   u_int8_t;
typedef _int64 int64_t;
#define popen(_Z,_X) _popen(_Z,_X)
#define pclose(_Z) _pclose(_Z)
#endif
#endif


typedef struct PFIND_INTERNAL_USE  // for datachk internal use
{
    char fname[128];
    char hdpat1[20],hdpat2[20];
	int bstep;  // read step in byte (default is 1) 
	FILE *fi;
} PFIND_INTERNAL_USE_T;


void pt_intuse_ini(PFIND_INTERNAL_USE_T *pt)
{
/*
   initialize DATACHK_INTERNAL_USE_T type variable

  */
	pt->fname[0]='\0';
	pt->hdpat1[0]='\0';
	pt->hdpat2[0]='\0';
	pt->bstep=1;
}

void patternconversion(char* hpat8, u_int32_t *pattern, u_int32_t *mask)
{
    /*
         Convert 8 digit HexAscii expression (eg. 8B????FF )
         to pattern ( 0x8B0000FF ) and mask ( FF0000FF )
    */
    int n,len;
    int t1,t2;
    u_int32_t w1,w2;
    len=strlen(hpat8);
    if(len > 8) len=8;
    w1=0; w2=0;
    for(n=0; n<len; n++){
        t1=0;
        t2=0;
        if(hpat8[n] >= 0x30 && hpat8[n] <= 0x39){  //  ASCII '0',....,'9' 
            t1=(int)(hpat8[n]-0x30);
            t2=0xF;
        } 
        if(hpat8[n] >= 0x41 && hpat8[n] <= 0x46){  //  ASCII 'A',....,'F' 
            t1=(int)(hpat8[n]-0x41)+10;
            t2=0xF;
        }
        if(hpat8[n] >= 0x61 && hpat8[n] <= 0x66){  //  ASCII 'a',....,'f' 
            t1=(int)(hpat8[n]-0x61)+10;
            t2=0xF;
        }
        //if(hpat8[n] == 0x3F){}  //  ASCII '?' 
        //if(hpat8[n] == 0x58){}  //  ASCII 'X' 
        //if(hpat8[n] == 0x78){}  //  ASCII 'x' 
        //if(hpat8[n] == 0x2A){}  //  ASCII '*' 
        w1=(w1<<4)|t1;
        w2=(w2<<4)|t2;
    }
    *pattern=w1;
    *mask=w2;
}


int pattern_find(PFIND_INTERNAL_USE_T *pt) 
{
    /* check K5 data (headder position only) */
	u_int32_t head32[2]; // to contain 8 bytes data 
	u_int16_t *head16; // pointer to head32 in 2 bytes unit  
	unsigned char *head8; // pointer to head32 in byte unit 
	int sa, bytesread;
	//int pos,posold;   // current pointer
	//int posbh;  // pointer at 1st of pattern found
	int64_t posbh;  // pointer at 1st of pattern found    2014-04-16
	int64_t pos,posold;   // current pointer              2014-04-16
	int i, ierr;
	int nkai;
    int nmax;        // 2018-01-08
    int non_stop;    // 2018-01-08
    char nyes;       // 2018-01-08

	u_int32_t pattern1,pattern2,mask1,mask2;

	head16=(u_int16_t*)head32;
	head8=(unsigned char*)head32;

	printf("Search Pattern Commanded  :   %s %s\n",pt->hdpat1,pt->hdpat2);
    patternconversion(pt->hdpat1, &pattern1, &mask1);
    patternconversion(pt->hdpat2, &pattern2, &mask2);

    printf("Pattern %08X %08X \n",pattern1,pattern2);
	printf("Mask    %08X %08X searching....\n",mask1,mask2);

	ierr=0;
	pos=0;
	posold=0;
	nkai=0;

    non_stop=0;
    nmax=50;

    if( (pt->fi  = fopen( pt->fname, "rb" )) == NULL ) {
           //printf("File %s Open Error! \n",K5.fname);
        printf("File %s not found! \n",pt->fname);
		ierr=-1;
        goto exitend;
    }
    printf("FILE : %s opened\n",pt->fname);
	printf("\n");


	printf("  No.    Pattern found     Position(bytes)  Difference(bytes)\n");


	// minimum read step is byte (8bit)

	if(pt->bstep < 1) pt->bstep=1;
    //bytesread = fread( head8, sizeof( unsigned char ), 7, pt->fi );
    bytesread = fread( head8, sizeof( unsigned char ), 8-pt->bstep, pt->fi );
    pos+=bytesread;  // pointer position is in bytes
    if( ferror( pt->fi) || bytesread==0 ){ // File EOF or read error
        i=fclose( pt->fi ) ;
		ierr=-1;
        goto exitend;
	}
    //printf("Debug %X  %X\n",head32,head32+1);  //debug

	while(1){
       /* pt->bstep byte data read */
       //bytesread = fread( head8+7, sizeof( unsigned char ), 1, pt->fi );
       bytesread = fread( head8+8-pt->bstep, sizeof( unsigned char ), pt->bstep, pt->fi );
       pos+=bytesread;  // pointer position is in bytes
       if( ferror( pt->fi) || bytesread==0 ){ // File EOF or read error
		   //printf("EOF\n");
           break;
	   }

	   posbh=pos-8; // 2013-03-06
	   // check pattern 
       if(((head32[0] & mask1)== (pattern1 & mask1))
	        &&((head32[1] & mask2)== (pattern2 & mask2))){
	      //sa=(int)(pos-posold);
	      sa=(int)(posbh-posold);  // 2013-03-06
          //printf("%5d Pattern found %08X %08X at pos=%12d  sa=%10d\n",
          //      nkai,head32[0],head32[1],pos,sa);
          //printf("%5d  %08X %08X   %12d %14d\n",
          printf("%5d  %08X %08X   %12lld %14d\n",     // 2014-04-16
				  nkai+1,head32[0],head32[1],(long long)posbh,sa);
          //posold=pos;
          posold=posbh;  // 2013-03-06
     	  nkai++;
          if(non_stop==0){                                                            //2018-01-08
             if((nkai != 0) && ((nkai % nmax)==0)){                                   //2018-01-08
                 printf("Hit Return key for NEXT (or \"N\"on-stop or \"Q\"uit) ");    //2018-01-08
                 nyes=getchar();  /* dummy read to accept just return */              //2018-01-08
                 if(nyes =='Q' || nyes=='q') break;                                   //2018-01-08
                 if(nyes =='N' || nyes=='n') non_stop=1;                              //2018-01-08
             }                                                                        //2018-01-08
          }                                                                           //2018-01-08
	   }
       // move pt->bstep byte
	   //memmove(head8, head8+1, 7);
	   memmove(head8, head8+pt->bstep, 8-pt->bstep);
	}
    i=fclose( pt->fi ) ;
 exitend:
    return ierr;
}

int main(int argc, char *argv[])
{
	int i;
    PFIND_INTERNAL_USE_T pt;
    if(argc < 2) {
       //strcpy(fname,dflt_fname);
       printf("%s %s\n",PROGRAM_NAME,PROGRAM_VERSION);
       printf("How to run: \n");
       printf("  pfind fname [HDpat1 [HDpat2]]\n");
       printf("\n");
       printf("    where  fname    -- file name\n");
	   printf("           HDpat1  ------ corrupted header pattern 1 in 8 digit Hexs or \'?\'\n");
       printf("                          where \'?\' is a wild card for one Hex.\n");
       //printf("                          (ex. FFFFFF??) (default is \'FFFFFFFF\')\n");
       printf("                          (ex. FFFFFF\?\?) (default is \'%s\')\n",DFLT_HDPAT1);  // '??)' is considered as trigraphs in gcc
	                                                                                            // to avoid it use \?\?
       printf("           HDpat2  ------ corrupted header pattern 3 in 8 digit Hexs or \'?\'\n");
       printf("                          where \'?\' is a wild card for one Hex.\n");
       //printf("                          (ex. 8C????FF) (default is \'8C??????\')\n");
       printf("                          (ex. 8B????FF for VSSP corrupted header,\n");
       printf("                               8C?????? for VSSP32 header\n");
       printf("                           default is \'%s\')\n",DFLT_HDPAT2);
       exit(0);
    }
    // initialize pt
    pt_intuse_ini(&pt);

	//pt.bstep=2;
    /* default corruped header pattern set for k51a */
    //strcpy(pt.hdpat1,"FFFFFFFF");  // default corrupted header pattern 1
    strcpy(pt.hdpat1,DFLT_HDPAT1);  // default corrupted header pattern 1
    //strcpy(pt.hdpat2,"8C??????");  // default corrupted header pattern 2
    strcpy(pt.hdpat2,DFLT_HDPAT2);  // default corrupted header pattern 2

	if (argc>=2){
       strcpy(pt.fname,*(argv+1));
    } 
    if(argc >=3){
       strcpy(pt.hdpat1,*(argv + 2 ));
    }
    if(argc >=4){
       strcpy(pt.hdpat2,*(argv + 3 ));
    }

    printf("************ %s  %s ************\n",PROGRAM_NAME,PROGRAM_VERSION);

	i=pattern_find(&pt);
    return 0;
}
